#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"
#include "oddebug.h"        /* This is also an example for using debug macros */

#define clear_bit(p,b) (p &= ~_BV(b))
#define set_bit(p,b)   (p |= _BV(b))

#define DATA_PORT   PINC
#define SELECT_PORT PORTB
#define SELECT_PIN  0

enum
{
  BTN_A,
  BTN_B,
  BTN_C,
  BTN_START,
  BTN_X,
  BTN_Y,
  BTN_Z,
  BTN_MODE
};
  
/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[43] = {
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x09, 0x05,                    // USAGE (Game Pad)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x09, 0x01,                    //   USAGE (Pointer)
  0xa1, 0x00,                    //   COLLECTION (Physical)
  0x09, 0x30,                    //     USAGE (X)
  0x09, 0x31,                    //     USAGE (Y)
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x26, 0xff, 0x00,              //     LOGICAL_MAXIMUM (255)
  0x75, 0x08,                    //   REPORT_SIZE (8)
  0x95, 0x02,                    //   REPORT_COUNT (2)
  0x81, 0x02,                    //   INPUT (Data,Var,Abs)
  0xc0,                          // END_COLLECTION
  0x05, 0x09,                    // USAGE_PAGE (Button)
  0x19, 0x01,                    //   USAGE_MINIMUM (Button 1)
  0x29, 0x08,                    //   USAGE_MAXIMUM (Button 8)
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    // REPORT_SIZE (1)
  0x95, 0x08,                    // REPORT_COUNT (8)
  0x81, 0x02,                    // INPUT (Data,Var,Abs)
  0xc0                           // END_COLLECTION
};

typedef struct
{
  uint8_t   dpad_x;
  uint8_t   dpad_y;
  uint8_t   buttons;
} report_t;

static report_t reportBuffer;
static uchar numButtons = 0; // 0=unknown, 3, or 6
static uchar idleRate;   /* repeat rate for keyboards, never used for mice */

static void readDB9()
{
  uchar data;
  reportBuffer.dpad_x = 128;
  reportBuffer.dpad_y = 128;
  reportBuffer.buttons = 0;
  
  // read d-pad, B, and C buttons  
  data = ~DATA_PORT; // invert data to make checking easier
  
  if (data & _BV(0)) // up
    reportBuffer.dpad_y = 0;
  else if (data & _BV(1)) // down
    reportBuffer.dpad_y = 255;

  if (data & _BV(2)) // left
    reportBuffer.dpad_x = 0;
  else if (data & _BV(3)) // right
    reportBuffer.dpad_x = 255;

  if (data & _BV(4))
    reportBuffer.buttons |= _BV(BTN_B);
  if (data & _BV(5))
    reportBuffer.buttons |= _BV(BTN_C);

  // pulse select low, read A and Start buttons
  clear_bit(SELECT_PORT, SELECT_PIN);
  _delay_us(20);
  data = ~DATA_PORT;
  
  if (data & _BV(4))
    reportBuffer.buttons |= _BV(BTN_A);
  if (data & _BV(5))
    reportBuffer.buttons |= _BV(BTN_START);
  
  set_bit(SELECT_PORT, SELECT_PIN);
  
  // For 6-button controller:
  // pulse the select line 2 more times,
  // after the second pulse, the controller responds with
  // X, Y, Z, and Mode
  if (numButtons != 3)
  {
    clear_bit(SELECT_PORT, SELECT_PIN);
    _delay_us(20);
    set_bit(SELECT_PORT, SELECT_PIN);
    _delay_us(20);
    clear_bit(SELECT_PORT, SELECT_PIN);
    _delay_us(20);
    
    // If this is a 6-button controller, it will respond
    // with data lines 0 and 1 pulled low after the third low pulse.
    // If either of these bits are set, this is a 3-button controller.
    /*if (numButtons == 0)
    {
      data = DATA_PORT;
      if (data & 0x3)
      {
        numButtons = 3; // turn off 6-button checking
        set_bit(SELECT_PORT, SELECT_PIN);
        return;
      }
      else
        numButtons = 6;
    }*/
    
    // Set the select line high to retrieve X, Y, Z, and Mode
    set_bit(SELECT_PORT, SELECT_PIN);
    _delay_us(20);
  
    data = ~DATA_PORT;
    
    if (data & _BV(0))
      reportBuffer.buttons |= _BV(BTN_Z);
    if (data & _BV(1))
      reportBuffer.buttons |= _BV(BTN_Y);
    if (data & _BV(2))
      reportBuffer.buttons |= _BV(BTN_X);
    if (data & _BV(3))
      reportBuffer.buttons |= _BV(BTN_MODE);
      
    // Pulse once more to reset the controller
    clear_bit(SELECT_PORT, SELECT_PIN);
    _delay_us(20);
    set_bit(SELECT_PORT, SELECT_PIN);
  }
}

/* ------------------------------------------------------------------------- */

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
  usbRequest_t    *rq = (void *)data;

  /* The following requests are never used. But since they are required by
   * the specification, we implement them in this example.
   */
  if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)    /* class request type */
  {
    DBG1(0x50, &rq->bRequest, 1);   /* debug output: print our request */
    if(rq->bRequest == USBRQ_HID_GET_REPORT)  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
    {
      /* we only have one report type, so don't look at wValue */
      usbMsgPtr = (void *)&reportBuffer;
      return sizeof(reportBuffer);
    }
    else if (rq->bRequest == USBRQ_HID_GET_IDLE)
    {
      usbMsgPtr = &idleRate;
      return 1;
    }
    else if(rq->bRequest == USBRQ_HID_SET_IDLE) 
    {
      idleRate = rq->wValue.bytes[1];
    }
  }
  else
  {
    /* no vendor specific requests implemented */
  }
  return 0;   /* default for not implemented requests: return no data back to host */
}

/* ------------------------------------------------------------------------- */

int main(void)
{
  uchar i;
  wdt_enable(WDTO_1S);

  /* Even if you don't use the watchdog, turn it off here. On newer devices,
   * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
   */
  DBG1(0x00, 0, 0);       /* debug output: main starts */
  /* RESET status: all port bits are inputs without pull-up.
   * That's the way we need D+ and D-. Therefore we don't need any
   * additional hardware initialization.
   */
  odDebugInit();
  usbInit();
  usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
  i = 0;
  while(--i)             /* fake USB disconnect for > 250 ms */
  {
    wdt_reset();
    _delay_ms(1);
  }
  usbDeviceConnect();
  
  /* set up DB9 pins */
  DDRC = 0x00; // six data pins, all inputs
  PORTC = 0xFF; // enable internal pull-ups
  DDRB |= _BV(SELECT_PIN); // select pin is an output
  set_bit(SELECT_PORT, SELECT_PIN); // select pin is normally high
  
  sei();
  DBG1(0x01, 0, 0);       /* debug output: main loop starts */
  for(;;)                /* main event loop */
  {
    DBG1(0x02, 0, 0);   /* debug output: main loop iterates */
    wdt_reset();
    usbPoll();
    if(usbInterruptIsReady())
    {
      /* called after every poll of the interrupt endpoint */
      readDB9();
      DBG1(0x03, 0, 0);   /* debug output: interrupt report prepared */
      usbSetInterrupt((void *)&reportBuffer, sizeof(reportBuffer));
    }
  }
  return 0;
}
